@@ -5,79 +5,14 @@
  */
 #include "cache.h"
 
-static int cache_name_compare(const char *name1, int len1, const char *name2, int len2)
-{
-	int len = len1 < len2 ? len1 : len2;
-	int cmp;
-
-	cmp = memcmp(name1, name2, len);
-	if (cmp)
-		return cmp;
-	if (len1 < len2)
-		return -1;
-	if (len1 > len2)
-		return 1;
-	return 0;
-}
-
-static int cache_name_pos(const char *name, int namelen)
-{
-	int first, last;
-
-	first = 0;
-	last = active_nr;
-	while (last > first) {
-		int next = (last + first) >> 1;
-		struct cache_entry *ce = active_cache[next];
-		int cmp = cache_name_compare(name, namelen, ce->name, ce->namelen);
-		if (!cmp)
-			return -next-1;
-		if (cmp < 0) {
-			last = next;
-			continue;
-		}
-		first = next+1;
-	}
-	return first;
-}
-
-static int remove_file_from_cache(char *path)
-{
-	int pos = cache_name_pos(path, strlen(path));
-	if (pos < 0) {
-		pos = -pos-1;
-		active_nr--;
-		if (pos < active_nr)
-			memmove(active_cache + pos, active_cache + pos + 1, (active_nr - pos - 1) * sizeof(struct cache_entry *));
-	}
-	return 0;
-}
-
-static int add_cache_entry(struct cache_entry *ce)
-{
-	int pos;
-
-	pos = cache_name_pos(ce->name, ce->namelen);
-
-	/* existing match? Just replace it */
-	if (pos < 0) {
-		active_cache[-pos-1] = ce;
-		return 0;
-	}
-
-	/* Make sure the array is big enough .. */
-	if (active_nr == active_alloc) {
-		active_alloc = alloc_nr(active_alloc);
-		active_cache = realloc(active_cache, active_alloc * sizeof(struct cache_entry *));
-	}
-
-	/* Add it in.. */
-	active_nr++;
-	if (active_nr > pos)
-		memmove(active_cache + pos + 1, active_cache + pos, (active_nr - pos - 1) * sizeof(ce));
-	active_cache[pos] = ce;
-	return 0;
-}
+/*
+ * Default to not allowing changes to the list of files. The
+ * tool doesn't actually care, but this makes it harder to add
+ * files to the revision control by mistake by doing something
+ * like "update-cache *" and suddenly having all the object
+ * files be revision controlled.
+ */
+static int allow_add = 0, allow_remove = 0;
 
 static int index_fd(const char *path, int namelen, struct cache_entry *ce, int fd, struct stat *st)
 {
@@ -122,6 +57,23 @@ static int index_fd(const char *path, int namelen, struct cache_entry *ce, int f
 	return write_sha1_buffer(ce->sha1, out, stream.total_out);
 }
 
+/*
+ * This only updates the "non-critical" parts of the directory
+ * cache, ie the parts that aren't tracked by GIT, and only used
+ * to validate the cache.
+ */
+static void fill_stat_cache_info(struct cache_entry *ce, struct stat *st)
+{
+	ce->ctime.sec = st->st_ctime;
+	ce->ctime.nsec = st->st_ctim.tv_nsec;
+	ce->mtime.sec = st->st_mtime;
+	ce->mtime.nsec = st->st_mtim.tv_nsec;
+	ce->st_dev = st->st_dev;
+	ce->st_ino = st->st_ino;
+	ce->st_uid = st->st_uid;
+	ce->st_gid = st->st_gid;
+}
+
 static int add_file_to_cache(char *path)
 {
 	int size, namelen;
@@ -131,8 +83,10 @@ static int add_file_to_cache(char *path)
 
 	fd = open(path, O_RDONLY);
 	if (fd < 0) {
-		if (errno == ENOENT)
-			return remove_file_from_cache(path);
+		if (errno == ENOENT) {
+			if (allow_remove)
+				return remove_file_from_cache(path);
+		}
 		return -1;
 	}
 	if (fstat(fd, &st) < 0) {
@@ -144,54 +98,108 @@ static int add_file_to_cache(char *path)
 	ce = malloc(size);
 	memset(ce, 0, size);
 	memcpy(ce->name, path, namelen);
-	ce->ctime.sec = st.st_ctime;
-	ce->ctime.nsec = st.st_ctim.tv_nsec;
-	ce->mtime.sec = st.st_mtime;
-	ce->mtime.nsec = st.st_mtim.tv_nsec;
-	ce->st_dev = st.st_dev;
-	ce->st_ino = st.st_ino;
+	fill_stat_cache_info(ce, &st);
 	ce->st_mode = st.st_mode;
-	ce->st_uid = st.st_uid;
-	ce->st_gid = st.st_gid;
 	ce->st_size = st.st_size;
 	ce->namelen = namelen;
 
 	if (index_fd(path, namelen, ce, fd, &st) < 0)
 		return -1;
 
-	return add_cache_entry(ce);
+	return add_cache_entry(ce, allow_add);
 }
 
-static int write_cache(int newfd, struct cache_entry **cache, int entries)
+static int match_data(int fd, void *buffer, unsigned long size)
 {
-	SHA_CTX c;
-	struct cache_header hdr;
-	int i;
+	while (size) {
+		char compare[1024];
+		int ret = read(fd, compare, sizeof(compare));
 
-	hdr.signature = CACHE_SIGNATURE;
-	hdr.version = 1;
-	hdr.entries = entries;
-
-	SHA1_Init(&c);
-	SHA1_Update(&c, &hdr, offsetof(struct cache_header, sha1));
-	for (i = 0; i < entries; i++) {
-		struct cache_entry *ce = cache[i];
-		int size = ce_size(ce);
-		SHA1_Update(&c, ce, size);
-	}
-	SHA1_Final(hdr.sha1, &c);
-
-	if (write(newfd, &hdr, sizeof(hdr)) != sizeof(hdr))
-		return -1;
-
-	for (i = 0; i < entries; i++) {
-		struct cache_entry *ce = cache[i];
-		int size = ce_size(ce);
-		if (write(newfd, ce, size) != size)
+		if (ret <= 0 || ret > size || memcmp(buffer, compare, ret))
 			return -1;
+		size -= ret;
+		buffer += ret;
 	}
 	return 0;
-}		
+}
+
+static int compare_data(struct cache_entry *ce)
+{
+	int match = -1;
+	int fd = open(ce->name, O_RDONLY);
+
+	if (fd >= 0) {
+		void *buffer;
+		unsigned long size;
+		char type[10];
+
+		buffer = read_sha1_file(ce->sha1, type, &size);
+		if (buffer) {
+			if (size == ce->st_size && !strcmp(type, "blob"))
+				match = match_data(fd, buffer, size);
+			free(buffer);
+		}
+		close(fd);
+	}
+	return match;
+}
+
+/*
+ * "refresh" does not calculate a new sha1 file or bring the
+ * cache up-to-date for mode/content changes. But what it
+ * _does_ do is to "re-match" the stat information of a file
+ * with the cache, so that you can refresh the cache for a
+ * file that hasn't been changed but where the stat entry is
+ * out of date.
+ *
+ * For example, you'd want to do this after doing a "read-tree",
+ * to link up the stat cache details with the proper files.
+ */
+static struct cache_entry *refresh_entry(struct cache_entry *ce)
+{
+	struct stat st;
+	struct cache_entry *updated;
+	int changed, size;
+
+	if (stat(ce->name, &st) < 0)
+		return NULL;
+
+	changed = cache_match_stat(ce, &st);
+	if (!changed)
+		return ce;
+
+	/*
+	 * If the length has changed, there's no point in trying
+	 * to refresh the entry - it's not going to match
+	 */
+	if (changed & (DATA_CHANGED | MODE_CHANGED))
+		return NULL;
+
+	if (compare_data(ce))
+		return NULL;
+
+	size = ce_size(ce);
+	updated = malloc(size);
+	memcpy(updated, ce, size);
+	fill_stat_cache_info(updated, &st);
+	return updated;
+}
+
+static void refresh_cache(void)
+{
+	int i;
+
+	for (i = 0; i < active_nr; i++) {
+		struct cache_entry *ce = active_cache[i];
+		struct cache_entry *new = refresh_entry(ce);
+
+		if (!new) {
+			printf("%s: needs update\n", ce->name);
+			continue;
+		}
+		active_cache[i] = new;
+	}
+}
 
 /*
  * We fundamentally don't like some paths: we don't want
@@ -224,6 +232,7 @@ inside:
 int main(int argc, char **argv)
 {
 	int i, newfd, entries;
+	int allow_options = 1;
 
 	entries = read_cache();
 	if (entries < 0) {
@@ -238,6 +247,26 @@ int main(int argc, char **argv)
 	}
 	for (i = 1 ; i < argc; i++) {
 		char *path = argv[i];
+
+		if (allow_options && *path == '-') {
+			if (!strcmp(path, "--")) {
+				allow_options = 0;
+				continue;
+			}
+			if (!strcmp(path, "--add")) {
+				allow_add = 1;
+				continue;
+			}
+			if (!strcmp(path, "--remove")) {
+				allow_remove = 1;
+				continue;
+			}
+			if (!strcmp(path, "--refresh")) {
+				refresh_cache();
+				continue;
+			}
+			usage("unknown option %s", path);
+		}
 		if (!verify_path(path)) {
 			fprintf(stderr, "Ignoring path %s\n", argv[i]);
 			continue;
